"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const ip = require("ip");
const constants_1 = require("../common/constants");
const util_1 = require("../common/util");
const net = require("net");
const smart_buffer_1 = require("smart-buffer");
class SocksClient extends events_1.EventEmitter {
    constructor(options) {
        super();
        this._options = Object.assign({}, options, { commandValue: util_1.getSocksCommandFromSocksCommandOption(options.command) });
        // todo: Normalize options (default options for nested objects)
        this._state = constants_1.SocksClientState.Created;
    }
    // Factory methods are mainly for CONNECT/CONNECT w/ chaining. (backwards compatible)
    // v2 adds promise support and connect chaining.
    static createConnection(options, callback) {
        // todo: throw error for non connect command.
        return new Promise((resolve, reject) => {
            const client = new SocksClient(options);
            client.connect();
            client.once('established', (info) => {
                if (typeof callback === 'function') {
                    callback(null, info);
                }
                resolve(info);
            });
            client.once('error', (err) => {
                if (typeof callback === 'function') {
                    callback(err);
                }
                reject(err);
            });
            client.once('close', () => {
                if (typeof callback === 'function') {
                    callback(null);
                }
                reject();
            });
        });
    }
    static createConnectionChain(options, callback) {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (options.randomizeChain) {
                util_1.shuffleArray(options.proxies);
            }
            /*
            proxy: {
              ipaddress: '104.236.216.222',
              port: 1081,
              type: 5
            },
      
            command: 'associate',
      
            destination: {
              //host: '104.54.207.254',
              //host: '84.19.186.186',
              host: '0.0.0.0',
              port: 0
            }*/
            for (let i = 0; i < options.proxies.length; i++) { }
        }));
    }
    /*
    var client = new TcpClient();
              var proxy = proxies[0];
  
              // Connect to first proxy
              await client.ConnectAsync(proxy.Endpoint.Address, proxy.Endpoint.Port);
  
              //Console.WriteLine("Local -> {0}", proxy);
  
              for(var i = 1; i <= proxies.Length; i++)
              {
                  var target = (i == proxies.Length) ? remoteserver : proxies[i].Endpoint;
  
                  if (proxy.Type == ProxyVersion.Socks4)
                      await NegotiateSocks4(client, target);
                  else
                      await NegotiateSocks5(client, target);
  
                  //Console.WriteLine("{0} -> {1}", proxies[i - 1], target);
              }
  
              return client;
    */
    static createUDPFrame(options) {
        const buff = new smart_buffer_1.SmartBuffer();
        buff.writeUInt16BE(0);
        buff.writeUInt8(options.frameNumber || 0);
        // IPv4/IPv6/Hostname
        if (net.isIPv4(options.remoteHost.host)) {
            buff.writeUInt8(constants_1.Socks5HostType.IPv4);
            buff.writeUInt32BE(ip.toLong(options.remoteHost.host));
        }
        else if (net.isIPv6(options.remoteHost.host)) {
            buff.writeUInt8(constants_1.Socks5HostType.IPv6);
            buff.writeBuffer(ip.toBuffer(options.remoteHost.host));
        }
        else {
            buff.writeUInt8(constants_1.Socks5HostType.Hostname);
            buff.writeUInt8(Buffer.byteLength(options.remoteHost.host));
            buff.writeString(options.remoteHost.host);
        }
        // Port
        buff.writeUInt16BE(options.remoteHost.port);
        // Data
        buff.writeBuffer(options.data);
        return buff.toBuffer();
    }
    static parseUDPFrame(data) {
        const buff = smart_buffer_1.SmartBuffer.fromBuffer(data);
        buff.readOffset = 2;
        const frameNumber = buff.readUInt8();
        const hostType = buff.readUInt8();
        let remoteHost;
        if (hostType === constants_1.Socks5HostType.IPv4) {
            remoteHost = ip.fromLong(buff.readUInt32BE());
        }
        else if (hostType === constants_1.Socks5HostType.IPv6) {
            remoteHost = ip.toString(buff.readBuffer(16));
        }
        else {
            remoteHost = buff.readString(buff.readUInt8());
        }
        const remotePort = buff.readUInt16BE();
        return {
            frameNumber,
            remoteHost: {
                host: remoteHost,
                port: remotePort,
            },
            data: buff.readBuffer(),
        };
    }
    connect(socket) {
        this._onDataReceived = (data) => this.onDataReceived(data);
        // Start timeout timer (defaults to 30 seconds)
        setTimeout(() => this.onEstablishedTimeout(), this._options.timeout || 30000);
        if (socket) {
            this._socket = socket;
        }
        else {
            this._socket = new net.Socket();
        }
        this._socket.once('close', (had_error) => this.onClose(had_error));
        this._socket.once('error', (err) => this.onError(err));
        this._socket.once('connect', () => this.onConnect());
        this._socket.on('data', this._onDataReceived);
        this._state = constants_1.SocksClientState.Connecting;
        if (socket) {
            this._socket.emit('connect');
        }
        else {
            this._socket.connect(this._options.proxy.port, this._options.proxy.ipaddress);
        }
        // Listens for instance 'established' event to remove internal data socket handlers.
        this.once('established', () => this.removeInternalSocketHandlers());
    }
    /**
     * Handles internal Socks timeout callback.
     * Note: If the Socks client is not BoundWaitingForConnection or Established, the connection will be closed.
     */
    onEstablishedTimeout() {
        if (this._state !== constants_1.SocksClientState.Established &&
            this._state !== constants_1.SocksClientState.BoundWaitingForConnection) {
            this._closeSocket('Proxy connection timed out');
        }
    }
    /**
     * Handles Socket connect event.
     */
    onConnect() {
        this._state = constants_1.SocksClientState.Connected;
        // Send initial handshake.
        if (this._options.proxy.type === 4) {
            this.sendSocks4InitialHandshake();
        }
        else {
            this.sendSocks5InitialHandshake();
        }
        this._state = constants_1.SocksClientState.SentInitialHandshake;
    }
    /**
     * Handles Socket data event.
     * @param data
     */
    onDataReceived(data) {
        // Sent initial handshake, waiting for response.
        if (this._state === constants_1.SocksClientState.SentInitialHandshake) {
            if (this._options.proxy.type === 4) {
                // Socks v4 only has one handshake response.
                this.handleSocks4FinalHandshakeResponse(data);
            }
            else {
                // Socks v5 has two handshakes.
                this.handleInitialSocks5HandshakeResponse(data);
            }
            // Sent auth request for Socks v5, waiting for response.
        }
        else if (this._state === constants_1.SocksClientState.SentAuthenication) {
            this.handleInitialSocks5AuthenticationHandshakeResponse(data);
            // Sent final Socks v5 handshake, waiting for final response.
        }
        else if (this._state === constants_1.SocksClientState.SentFinalHandshake) {
            this.handleSocks5FinalHandshakeResponse(data);
            // Socks BIND established. Waiting for remote connection via proxy.
        }
        else if (this._state === constants_1.SocksClientState.BoundWaitingForConnection) {
            if (this._options.proxy.type === 4) {
                this.handleSocks4IncomingConnectionResponse(data);
            }
            else {
                this.handleSocks5IncomingConnectionResponse(data);
            }
        }
    }
    /**
     * Handles Socket close event.
     * @param had_error
     */
    onClose(had_error) {
        this._closeSocket();
    }
    /**
     * Handles Socket error event.
     * @param err
     */
    onError(err) {
        this._closeSocket(err.message);
    }
    /**
     * Removes internal 'data' event listener on underlying Socket.
     */
    removeInternalSocketHandlers() {
        this._socket.removeListener('data', this._onDataReceived);
    }
    /**
     * Closes and destroys the underlying Socket.
     *
     * Note: Either only the 'close' or 'error' event will be emitted in a SocksClient lifetime. Never both.
     * @param err Optional error message to include in error event.
     */
    _closeSocket(err) {
        if (this._state !== constants_1.SocksClientState.Error &&
            this._state !== constants_1.SocksClientState.Closed) {
            this._socket.destroy(err);
            if (err) {
                this._state = constants_1.SocksClientState.Error;
                this.emit('error', new util_1.SocksClientError(err, this._options));
            }
            else {
                this._state = constants_1.SocksClientState.Closed;
                this.emit('close');
            }
        }
    }
    /**
     * Sends initial Socks v4 handshake request.
     */
    sendSocks4InitialHandshake() {
        const buff = new smart_buffer_1.SmartBuffer();
        buff.writeUInt8(0x04);
        buff.writeUInt8(this._options.commandValue);
        buff.writeUInt16BE(this._options.destination.port);
        // Socks 4 (IPv4)
        if (net.isIPv4(this._options.destination.host)) {
            buff.writeBuffer(ip.toBuffer(this._options.destination.host));
            buff.writeStringNT(this._options.proxy.userId || '');
            // Socks 4a (hostname)
        }
        else {
            buff.writeUInt8(0x00);
            buff.writeUInt8(0x00);
            buff.writeUInt8(0x00);
            buff.writeUInt8(0x01);
            buff.writeStringNT(this._options.proxy.userId || '');
            buff.writeStringNT(this._options.destination.host);
        }
        this._socket.write(buff.toBuffer());
    }
    /**
     * Handles Socks v4 handshake response.
     * @param data
     */
    handleSocks4FinalHandshakeResponse(data) {
        if (data.length < 8) {
            this._closeSocket('Received invalid Socks4 handshake response');
        }
        else if (data[1] !== constants_1.Socks4Response.Granted) {
            this._closeSocket(`Server rejected connection (${constants_1.Socks4Response[data[1]]})`);
        }
        else {
            // Bind response
            if (this._options.commandValue === constants_1.SocksCommand.Bind) {
                const buff = smart_buffer_1.SmartBuffer.fromBuffer(data);
                buff.readOffset = 2;
                const remoteHostInfo = {
                    port: buff.readUInt16BE(),
                    host: ip.fromLong(buff.readUInt32BE()),
                };
                // If host is 0.0.0.0, set to proxy host.
                if (remoteHostInfo.host === '0.0.0.0') {
                    remoteHostInfo.host = this._options.proxy.ipaddress;
                }
                this._state = constants_1.SocksClientState.BoundWaitingForConnection;
                this.emit('bound', { socket: this._socket, remoteHostInfo });
                // Connect response
            }
            else {
                this._state = constants_1.SocksClientState.Established;
                this._socket.pause();
                this.emit('established', { socket: this._socket });
            }
        }
    }
    /**
     * Sends initial Socks v5 handshake request.
     */
    sendSocks5InitialHandshake() {
        const buff = new smart_buffer_1.SmartBuffer();
        buff.writeUInt8(0x05);
        buff.writeUInt8(2);
        buff.writeUInt8(constants_1.Socks5Auth.NoAuth);
        buff.writeUInt8(constants_1.Socks5Auth.UserPass);
        this._socket.write(buff.toBuffer());
        this._state = constants_1.SocksClientState.SentInitialHandshake;
    }
    /**
     * Handles initial Socks v5 handshake response.
     * @param data
     */
    handleInitialSocks5HandshakeResponse(data) {
        if (data.length !== 2) {
            this._closeSocket('Negotiation Error');
        }
        else if (data[0] !== 0x05) {
            this._closeSocket('Negotiation Error (invalid socks version)');
        }
        else if (data[1] === 0xff) {
            this._closeSocket('Negotiation Error (no accepted authentication type)');
        }
        else {
            // If selected Socks v5 auth method is no auth, send final handshake request.
            if (data[1] === constants_1.Socks5Auth.NoAuth) {
                this.sendSocks5CommandRequest();
                // If selected Socks v5 auth method is user/password, send auth handshake.
            }
            else if (data[1] === constants_1.Socks5Auth.UserPass) {
                this.sendSocks5Authentication();
            }
            else {
                this._closeSocket('Negotiation Error (unknown authentication type)');
            }
        }
    }
    /**
     * Sends Socks v5 auth handshake.
     *
     * Note: No auth and user/pass are currently supported.
     */
    sendSocks5Authentication() {
        const buff = new smart_buffer_1.SmartBuffer();
        buff.writeUInt8(0x01);
        buff.writeUInt8(Buffer.byteLength(this._options.proxy.userId));
        buff.writeString(this._options.proxy.userId);
        buff.writeUInt8(Buffer.byteLength(this._options.proxy.password));
        buff.writeString(this._options.proxy.password);
        this._socket.write(buff.toBuffer());
        this._state = constants_1.SocksClientState.SentAuthenication;
    }
    /**
     * Handles Socks v5 auth handshake response.
     * @param data
     */
    handleInitialSocks5AuthenticationHandshakeResponse(data) {
        this._state = constants_1.SocksClientState.ReceivedAuthenticationResponse;
        if (data.length === 2 && data[1] === 0x00) {
            this.sendSocks5CommandRequest();
        }
        else {
            this._closeSocket('Negotiation Error (authentication failed)');
        }
    }
    /**
     * Sends Socks v5 final handshake request.
     */
    sendSocks5CommandRequest() {
        const buff = new smart_buffer_1.SmartBuffer();
        buff.writeUInt8(0x05);
        buff.writeUInt8(this._options.commandValue);
        buff.writeUInt8(0x00);
        // ipv4, ipv6, domain?
        if (net.isIPv4(this._options.destination.host)) {
            buff.writeUInt8(0x01);
            buff.writeBuffer(ip.toBuffer(this._options.destination.host));
        }
        else if (net.isIPv6(this._options.destination.host)) {
            buff.writeUInt8(0x04);
            buff.writeBuffer(ip.toBuffer(this._options.destination.host));
        }
        else {
            buff.writeUInt8(0x03);
            buff.writeUInt8(this._options.destination.host.length);
            buff.writeString(this._options.destination.host);
        }
        buff.writeUInt16BE(this._options.destination.port);
        this._socket.write(buff.toBuffer());
        this._state = constants_1.SocksClientState.SentFinalHandshake;
    }
    /**
     * Handles Socks v5 final handshake response.
     * @param data
     */
    handleSocks5FinalHandshakeResponse(data) {
        if (data.length < 4) {
            this._closeSocket('Negotiation Error');
        }
        else if (data[0] !== 0x05 || data[1] !== constants_1.Socks5Response.Granted) {
            this._closeSocket('Negotiation Error');
        }
        else {
            this._state = constants_1.SocksClientState.ReceivedFinalResponse;
            if (this._options.commandValue === constants_1.SocksCommand.Connect) {
                this._state = constants_1.SocksClientState.Established;
                this.emit('established', { socket: this._socket });
            }
            else {
                // Read address type
                const buff = smart_buffer_1.SmartBuffer.fromBuffer(data);
                buff.readOffset = 3;
                const addressType = buff.readUInt8();
                let remoteHostInfo;
                // IPv4
                if (addressType === 0x01) {
                    remoteHostInfo = {
                        host: ip.fromLong(buff.readUInt32BE()),
                        port: buff.readUInt16BE(),
                    };
                    // If given host is 0.0.0.0, assume remote proxy ip instead.
                    if (remoteHostInfo.host === '0.0.0.0') {
                        remoteHostInfo.host = this._options.proxy.ipaddress;
                    }
                    // Hostname
                }
                else if (addressType === 0x03) {
                    const hostLength = buff.readUInt8();
                    remoteHostInfo = {
                        host: buff.readString(hostLength),
                        port: buff.readUInt16BE(),
                    };
                    // IPv6
                }
                else if (addressType === 0x04) {
                    remoteHostInfo = {
                        host: ip.toString(buff.readBuffer(16)),
                        port: buff.readUInt16BE(),
                    };
                }
                /* If using BIND, the Socks client is now in BoundWaitingForConnection state.
                   This means that the remote proxy server is waiting for a remote connection to the bound port. */
                if (this._options.commandValue === constants_1.SocksCommand.Bind) {
                    this._state = constants_1.SocksClientState.BoundWaitingForConnection;
                    this.emit('bound', { socket: this._socket, remoteHostInfo });
                    /*
                    If using Associate, the Socks client is now Established. And the proxy server is now accepting UDP packets at the
                    given bound port. This initial Socks TCP connection must remain open for the UDP relay to continue to work.
                  */
                }
                else if (this._options.commandValue === constants_1.SocksCommand.Associate) {
                    this._state = constants_1.SocksClientState.Established;
                    this.emit('established', { socket: this._socket, remoteHostInfo });
                }
            }
        }
    }
    handleSocks5IncomingConnectionResponse(data) {
        if (data.length < 4) {
            this._closeSocket('Negotiation Error');
        }
        else if (data[0] !== 0x05 || data[1] !== constants_1.Socks5Response.Granted) {
            this._closeSocket('Negotiation Error');
        }
        else {
            // <Buffer 05 00 00 01 68 ec d8 de 8b ac>
            // Read address type
            const buff = smart_buffer_1.SmartBuffer.fromBuffer(data);
            buff.readOffset = 3;
            const addressType = buff.readUInt8();
            let remoteHostInfo;
            // IPv4
            if (addressType === 0x01) {
                remoteHostInfo = {
                    host: ip.fromLong(buff.readUInt32BE()),
                    port: buff.readUInt16BE(),
                };
                if (remoteHostInfo.host === '0.0.0.0') {
                    remoteHostInfo.host = this._options.proxy.ipaddress;
                }
                // Hostname
            }
            else if (addressType === 0x03) {
                const hostLength = buff.readUInt8();
                remoteHostInfo = {
                    host: buff.readString(hostLength),
                    port: buff.readUInt16BE(),
                };
                // IPv6
            }
            else if (addressType === 0x04) {
                remoteHostInfo = {
                    host: ip.toString(buff.readBuffer(16)),
                    port: buff.readUInt16BE(),
                };
            }
            this._state = constants_1.SocksClientState.Established;
            this.emit('established', { socket: this._socket, remoteHostInfo });
        }
    }
    handleSocks4IncomingConnectionResponse(data) {
        if (data.length < 8) {
            this._closeSocket('Received invalid incoming connection response');
        }
        else if (data[1] !== constants_1.Socks4Response.Granted) {
            this._closeSocket(`Server rejected incoming bound connection (${constants_1.Socks4Response[data[1]]})`);
        }
        else {
            const buff = smart_buffer_1.SmartBuffer.fromBuffer(data);
            buff.readOffset = 2;
            const remoteHostInfo = {
                port: buff.readUInt16BE(),
                host: ip.fromLong(buff.readUInt32BE()),
            };
            this._state = constants_1.SocksClientState.Established;
            this._socket.pause();
            this.emit('established', { socket: this._socket, remoteHostInfo });
        }
    }
}
exports.SocksClient = SocksClient;
//# sourceMappingURL=socks-client.js.map